
import java.util.ArrayList;
/**
 * Verwaltet einen ungerichteten, gewichteten Graphen mittels Adjazenzmatrix
 * 
 * @author Albert Wiedemann 
 * @version 1.0
 */
class GraphMatrix
{
    /** Feld der Knoten des Graphen */
    private ArrayList<Knoten> knoten;   
    /** 2-dim Feld der Adjazenzmatrix */
    private ArrayList<ArrayList<Integer>> matrix;
    /** Die Warteliste (ToDo-Liste) */
    private ArrayList<Integer> warteliste;
    /** Feld der fertigen Knoten */
    private ArrayList<Integer> fertigeKnoten;

    /**
     * Baut die Datenstruktur auf
     */
    GraphMatrix ()
    {
        knoten = new ArrayList<Knoten>();
        matrix = new ArrayList<ArrayList<Integer>>();
        warteliste = new ArrayList<Integer>();
        fertigeKnoten = new ArrayList<Integer>();
    }

    /**
     * Einfügen eines neuen Knoten in den Graphen
     * @param bezeichner Bezeichner des neuen Knotens, der dem Graphen hinzugefügt wird.
     */
    void KnotenEinfügen(String bezeichner)
    {
        knoten.add(new Knoten(bezeichner));
        for (int index = 0; index < matrix.size(); index++)
        {
            matrix.get(index).add(new Integer(-1));
        }
        matrix.add(new ArrayList<Integer>());
        for (int index = 0; index < matrix.size(); index++)
        {
            matrix.get(matrix.size()-1).add(new Integer(-1));
        }
    }

    /**
     * Gibt den Knoten mit dem angegebenen Bezeichner zurück.
     * Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird null zurückgegeben
     * @param bezeichner Bezeichner des Knoten der gesucht wird
     * @return Referenz auf das Knotenobjekt oder null
     */   
    private Knoten KnotenGeben(String bezeichner)
    {
        for (Knoten k: knoten)
        {
            if (k.BezeichnerGeben().equals(bezeichner))
            {
                return k;
            }
        }
        return null;
    }

    /**
     * Gibt die interne Nummer des Knotens
     * Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird -1 zurückgegeben
     * @param bezeichner Bezeichner des Knoten der gesucht wird
     * @return Indexnummer des Knotens im Knotenfeld; 0 <= res <= knoten.size()-1 bzw. -1
     */   
    int KnotenNummerGeben(String bezeichner)
    {
        for (int index=0; index < knoten.size(); index++)
        {
            if (knoten.get(index).BezeichnerGeben().equals(bezeichner))
            {
                return index;
            }
        }
        return -1;
    }

    /**
     * Gibt die Bezeichnung eines Knotens mit der internen Knotennummer
     * @param Indexnummer des Knotens im Knotenarray; 0<= x <= knoten.size()
     * @return Bezeichner des Knotens
     */   
    String KnotenBezeichnerGeben(int knotenNummer)
    {
        if ((knotenNummer < knoten.size()) && (knotenNummer >= 0))
        {
            return knoten.get(knotenNummer).BezeichnerGeben();
        }
        else
        {
            return "";
        }
    }

    /**
     * Einfügen einer Kante in den Graphen
     * Eine Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt und hat eine Gewichtung
     * @param von Bezeichner des Anfangsknotens
     * @param nach Bezeichner des Endknotens
     * @param gewichtung Gewichtung der Kante als Ganzzahl
     */
    void KanteEinfügen(String von, String nach, int gewichtung)
    {
        int vonNummer, nachNummer;
        vonNummer = KnotenNummerGeben(von);
        nachNummer = KnotenNummerGeben(nach);
        if ((vonNummer!=-1) && (nachNummer!=-1) && (vonNummer!=nachNummer))
        {
            matrix.get(vonNummer).set(nachNummer, gewichtung);
            matrix.get(nachNummer).set(vonNummer, gewichtung);
        }
    }


    /**
     * Gibt die Anzahl der Knoten des Graphen zurück
     * @return  Anzahl der Knoten
     */   
    int KnotenAnzahlgeben()
    {
        return knoten.size();
    }

    /**
     * Gibt die Gewichtung einer Kante zurück
     * Die Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt
     * @param von Bezeichner des Anfangsknotens
     * @param nach Bezeichner des Endknotens
     * @return Gewichtung der Kante
     */ 
    int KanteGewichtGeben(String von, String nach)
    {
        int vonNummer, nachNummer;

        vonNummer = KnotenNummerGeben(von);
        nachNummer = KnotenNummerGeben(nach);
        if ((vonNummer!=-1) && (nachNummer!=-1))
        {
            return matrix.get(vonNummer).get(nachNummer);
        }
        else
        {
            return -1;
        }
    }

    /**
     * Löscht die Kanten und Knoten des Graphen
     * Die Anzeige wird auch gelöscht
     */
    void ZurückSetzen()
    {
        knoten.clear();
        matrix.clear();
    }

    /**
     * Führt den Dijkstra-Algorithmus vom Startknoten zum Zielknoten durch
     * @param startKnoten die Nummer des Startknotens
     * @param zielKnoten die Nummer des Zielknotens
     */
    void DijkstraAusführen(int startKnoten, int zielKnoten)
    {
        //Vorbelegung
        warteliste.clear();
        fertigeKnoten.clear();
        knoten.get(startKnoten).LängeSetzen(0);
        knoten.get(startKnoten).VorgängerSetzen(-1);
        //es geht los mit dem Startknoten
        int aktuellerKnoten = startKnoten;
        //solange Ziel nicht erreicht...
        while ((aktuellerKnoten != zielKnoten) && (aktuellerKnoten != -1))
        {
            //untersuche alle Knoten
            for (int nummer = 0; nummer < knoten.size(); nummer += 1)
            {
                //suche den Eintrag in der Adjazenzmatrix mit den Koordinaten (aktuellerKnoten|nummer)
                int längeNeu = matrix.get(aktuellerKnoten).get(nummer);
                //falls es eine Kante gibt und der Knoten noch nicht fertig ist...
                if ((längeNeu > 0) && (! fertigeKnoten.contains(nummer)))
                {
                    //überprüfe, ob der Knoten bereits einen Vorgänger hat
                    if ( warteliste.contains(nummer))
                    {
                        // überprüfe, ob es über den aktuellen Knoten eine kürzere Verbindung gibt als die bisher gefundene
                        if (knoten.get(aktuellerKnoten).LängeGeben() + längeNeu < knoten.get(nummer).LängeGeben())
                        {
                            //falls ja, aktualisiere die Länge und den Vorgänger des Knoten
                            knoten.get(nummer).LängeSetzen(knoten.get(aktuellerKnoten).LängeGeben() + längeNeu);
                            knoten.get(nummer).VorgängerSetzen(aktuellerKnoten);
                        }
                    }
                    else
                    {
                        //falls der Knoten bisher keinen Vorgänger hat, wird er in die Warteliste aufgenommen. 
                        warteliste.add(nummer);
                        knoten.get(nummer).VorgängerSetzen(aktuellerKnoten);
                        knoten.get(nummer).LängeSetzen(knoten.get(aktuellerKnoten).LängeGeben() + längeNeu);

                    }
                }
            }
            //der Knoten ist fertig
            fertigeKnoten.add(aktuellerKnoten);
            
            //Ausgabe des Zwischenstandes als Tabelle wie im Schulbuch
            System.out.print(("Ort"+"                     ").substring(0,20));
            System.out.print(("kommt von"+"                     ").substring(0,20));
            System.out.print(("Distanz zum Start"+"                     ").substring(0,20));
            System.out.println(("bearbeitet"+"                     ").substring(0,20));
            for (Knoten k: knoten)
            {
                System.out.print((k.BezeichnerGeben()+"                     ").substring(0,20));
                if(k.VorgängerGeben()!=-1)
                {
                    System.out.print((knoten.get(k.VorgängerGeben()).BezeichnerGeben()+"                     ").substring(0,20));
                }
                else
                {
                    System.out.print("kein Vorgänger                     ".substring(0,20));
                }
                System.out.print((k.LängeGeben()+"                     ").substring(0,20));
 
                if (fertigeKnoten.contains(KnotenNummerGeben(k.BezeichnerGeben())))
                {
                    System.out.println(("ok"+"                     ").substring(0,20));
                }
                else
                {
                    System.out.println("");
                }
            }
            System.out.println("");
            System.out.println("");

            //nun wird der nächste Knoten gesucht: ein bisher nicht besuchter Knoten mit minimaler Distanz zum Startknoten
            int indexOptimalerKnoten = 0;
            for (int nummer = 1; nummer < warteliste.size(); nummer += 1)
            {
                if (knoten.get(warteliste.get(indexOptimalerKnoten)).LängeGeben() >  knoten.get(warteliste.get(nummer)).LängeGeben())
                {
                    indexOptimalerKnoten = nummer;
                }
            }
            //falls noch Knoten vorhanden sind, wird der gerade gesuchte Knoten entfernt aus der Warteliste und als 
            //aktueller Knoten für die nächste Runde festgelegt. 
            if (warteliste.size() > 0)
            {
                aktuellerKnoten = warteliste.get(indexOptimalerKnoten);
                warteliste.remove(indexOptimalerKnoten);
            }
            else
            {
                aktuellerKnoten = -1;
            }
        }
        //Ausgabe des kürzesten Weges
        if (aktuellerKnoten != -1)
        {
            System.out.println("Der Weg hat die Länge " + knoten.get(aktuellerKnoten).LängeGeben());
            ArrayList<String> namen = new ArrayList<String>();
            while (aktuellerKnoten != -1)
            {
                namen.add(knoten.get(aktuellerKnoten).BezeichnerGeben());
                aktuellerKnoten = knoten.get(aktuellerKnoten).VorgängerGeben();
            }
            for (int index = namen.size() - 1; index >= 0; index -= 1)
            {
                System.out.println(namen.get(index));
            }
        }
        else
        {
            System.out.println("Ziel nicht erreichbar");
        }
    }
}
